@@ -38,7 +38,11 @@ iso_mount (long cons_dev, long p_offset, long quiet)
 	return iso_read_super(NULL, quiet);
 }
 
-
+static const char *
+iso_readdir(int fd, int rewind)
+{
+	return NULL;
+}
 
 const struct bootfs iso = {
   -1,	/* isofs is not partitioned */
@@ -46,5 +50,7 @@ const struct bootfs iso = {
   iso_mount,
   iso_open,
   iso_bread,
-  iso_close
+  iso_close,
+  iso_readdir,
+  iso_fstat
 };
