@@ -7,13 +7,14 @@
 # define fd_set kernel_fd_set
 #endif
 
-#include <linux/types.h>
+#include <asm/stat.h>
 
 extern int  iso_read_super (void * data, int quiet);
 extern int  iso_open (const char * filename);
 extern int  iso_bread (int fd, long blkno, long nblks, char * buffer);
 extern void iso_close (int fd);
 extern long iso_map (int fd, long block);
-extern size_t iso_fsize (int fd);
+extern int  iso_fstat (int fd, struct stat * buf);
+extern int isonum_711 (char * p);
 
 #endif /* isolib_h */
