@@ -6,11 +6,15 @@
  * functionality to the Linux bootstrapper.  All we can do is
  * open and read files... but that's all we need 8-)
  */
+#ifndef TESTING
+#  include <linux/string.h>
+#endif
 #include <linux/kernel.h>
 #include <linux/config.h>
-#include <linux/string.h>
+#include <linux/types.h>
 #include <linux/stat.h>
 #include <linux/kdev_t.h>
+#include <asm/page.h>
 
 #include <iso.h>
 #include <isolib.h>
@@ -45,6 +49,8 @@ static char big_data_block[2048];
 # define S_IRUGO	(S_IRUSR|S_IRGRP|S_IROTH)
 #endif
 
+extern long iso_dev_read (void * buf, long offset, long size);
+
 
 int
 isonum_711 (char * p)
@@ -113,22 +119,38 @@ isonum_733 (char * p)
 
 
 static int
-iso_breadi (long blkno, long nblks, char * buffer)
+iso_bmap (struct iso_inode *inode, int block)
 {
-	extern long iso_dev_read (void * buf, long offset, long size);
-	long offset, size;
-
-#ifdef DEBUG_ISO
-	printf("iso_breadi: reading blocks %ld..%ld\n",
-	       blkno, blkno + nblks - 1);
-#endif
-	offset = blkno * sb.s_blocksize;
-	size   = nblks * sb.s_blocksize;
-	if (iso_dev_read(buffer, offset, size) != size) {
-		printf("iso_breadi: read error on block %ld\n", blkno);
-		return -1;
+	if (block < 0) {
+		printf("iso_bmap: block<0");
+		return 0;
 	}
-	return 0;
+	return (inode->i_first_extent >> sb.s_blocksize_bits) + block;
+}
+
+static int
+iso_breadi (struct iso_inode *ip, long blkno, long nblks, char * buffer)
+{
+	long i_size, abs_blkno;
+
+	/* do some error checking */
+	if (!ip || !ip->i_first_extent)
+		return -1;
+
+	i_size = ((struct inode_table_entry *) ip)->size;
+	/* as in ext2.c - cons_read() doesn't really cope well with
+           EOF conditions - actually it should be fixed */
+	if ((blkno+nblks) * sb.s_blocksize > i_size)
+		nblks = ((i_size + sb.s_blocksize)
+			 / sb.s_blocksize) - blkno;
+
+	/* figure out which iso block number(s) we're being asked for */
+	abs_blkno = iso_bmap(ip, blkno);
+	if (!abs_blkno)
+		return -1;
+	/* now try and read them (easy since ISO files are continguous) */
+	return iso_dev_read(buffer, abs_blkno * sb.s_blocksize,
+			    nblks * sb.s_blocksize);
 }
 
 
@@ -148,7 +170,6 @@ iso_iput (struct iso_inode *ip)
 	itp->free = 1;
 }
 
-
 /*
  * Read the specified inode from the disk and return it to the user.
  * Returns NULL if the inode can't be read...
@@ -185,11 +206,8 @@ iso_iget (int ino)
 	}
 
 	block = ino >> sb.s_blocksize_bits;
-#ifdef DEBUG_ISO
-	printf("iso_inode: reading ino %d, block %d at offset %d\n",
-	       ino, block, sb.s_blocksize * ino);
-#endif
-	if (iso_breadi(block, 1, data_block) < 0) {
+	if (iso_dev_read(data_block, block * sb.s_blocksize, sb.s_blocksize)
+	    != sb.s_blocksize) {
 		printf("iso9660: unable to read i-node block");
 		return NULL;
 	}
@@ -205,11 +223,15 @@ iso_iget (int ino)
 		frag1 = sb.s_blocksize - offset;
 		cpnt = big_data_block;
 		memcpy(cpnt, data_block + offset, frag1);
-		if (iso_breadi(++block, 1, data_block) < 0) {
+		offset += *pnt - sb.s_blocksize; /* DUH! pnt would get
+                                                    wiped out by the
+                                                    iso_dev_read here. */
+		if (iso_dev_read(data_block, ++block * sb.s_blocksize,
+				 sb.s_blocksize)
+		    != sb.s_blocksize) {
 			printf("unable to read i-node block");
 			return NULL;
 		}
-		offset += *pnt - sb.s_blocksize;
 		memcpy((char *)cpnt+frag1, data_block, offset);
 		pnt = ((unsigned char *) cpnt);
 		raw_inode = ((struct iso_directory_record *) pnt);
@@ -302,14 +324,6 @@ iso_iget (int ino)
 		break;
 	}
 
-#if 0
-	/* Now test for possible Rock Ridge extensions which will
-	   override some of these numbers in the inode structure. */
-
-	if (!high_sierra)
-		parse_rock_ridge_inode(raw_inode, inode);
-#endif
-
 	/* keep our inode table correct */
 	itp->free = 0;
 	itp->inumber = ino;
@@ -321,10 +335,10 @@ iso_iget (int ino)
 
 /*
  * ok, we cannot use strncmp, as the name is not in our data space.
- * Thus we'll have to use isofs_match. No big problem. Match also makes
+ * Thus we'll have to use iso_match. No big problem. Match also makes
  * some sanity tests.
  *
- * NOTE! unlike strncmp, isofs_match returns 1 for success, 0 for failure.
+ * NOTE! unlike strncmp, iso_match returns 1 for success, 0 for failure.
  */
 static int
 iso_match (int len, const char *name, const char *compare, int dlen)
@@ -333,7 +347,8 @@ iso_match (int len, const char *name, const char *compare, int dlen)
 		return 0;
 
 #ifdef DEBUG_ISO
-	printf("iso_match: comparing %s with %s\n", name, compare);
+	printf("iso_match: comparing %d chars of %s with %s\n",
+	       dlen, name, compare);
 #endif
 
 	/* check special "." and ".." files */
@@ -354,14 +369,84 @@ iso_match (int len, const char *name, const char *compare, int dlen)
 }
 
 
-static int
-iso_bmap (struct iso_inode *inode, int block)
+/* get_rock_ridge_filename() from fs/isofs/rock.c in the Linux kernel,
+   (c) 1992, 1993 Eric Youngdale */
+#define SIG(A,B) ((A << 8) | B)
+
+int get_rock_ridge_filename(struct iso_directory_record * de,
+			    char * retname,
+			    struct iso_inode * inode)
 {
-	if (block < 0) {
-		printf("iso_bmap: block<0");
-		return 0;
+  int len;
+  unsigned char * chr;
+  int retnamlen = 0, truncate=0;
+
+  /* No rock ridge? well then... */
+  if (!sb.s_rock) return 0;
+  *retname = 0;
+
+  len = sizeof(struct iso_directory_record) + de->name_len[0];
+  if (len & 1) len++;
+  chr = ((unsigned char *) de) + len;
+  len = *((unsigned char *) de) - len;
+  {
+    struct rock_ridge * rr;
+    int sig;
+    
+    while (len > 1){ /* There may be one byte for padding somewhere */
+      rr = (struct rock_ridge *) chr;
+      if (rr->len == 0) goto out; /* Something got screwed up here */
+      sig = (chr[0] << 8) + chr[1];
+      chr += rr->len; 
+      len -= rr->len;
+
+      switch(sig){
+      case SIG('R','R'):
+	if((rr->u.RR.flags[0] & RR_NM) == 0) goto out;
+	break;
+      case SIG('S','P'):
+        if (rr->u.SP.magic[0] != 0xbe ||
+	    rr->u.SP.magic[1] != 0xef)
+	  goto out;
+	break;
+      case SIG('C','E'):
+	printf("sorry, continued records not supported\n");
+	goto out;
+	break;
+      case SIG('N','M'):
+	if (truncate) break;
+        /*
+	 * If the flags are 2 or 4, this indicates '.' or '..'.
+	 * We don't want to do anything with this, because it
+	 * screws up the code that calls us.  We don't really
+	 * care anyways, since we can just use the non-RR
+	 * name.
+	 */
+	if (rr->u.NM.flags & 6) {
+	  break;
 	}
-	return (inode->i_first_extent >> sb.s_blocksize_bits) + block;
+
+	if (rr->u.NM.flags & ~1) {
+	  printf("Unsupported NM flag settings (%d)\n",rr->u.NM.flags);
+	  break;
+	};
+	if((strlen(retname) + rr->len - 5) >= 254) {
+	  truncate = 1;
+	  break;
+	};
+	strncat(retname, rr->u.NM.name, rr->len - 5);
+	retnamlen += rr->len - 5;
+	break;
+      case SIG('R','E'):
+	return -1;
+      default:
+	break;
+      }
+    };
+  }
+  return retnamlen; /* If 0, this file did not have a NM field */
+ out:
+  return 0;
 }
 
 
@@ -381,8 +466,7 @@ iso_find_entry (struct iso_inode *dir, const char *name, int namelen,
 	void * cpnt = NULL;
 	unsigned int old_offset;
 	unsigned int backlink;
-	int dlen, rrflag, match, i;
-	char * dpnt;
+	int dlen, match, i;
 	struct iso_directory_record * de;
 	char c;
 	struct inode_table_entry *itp = (struct inode_table_entry *) dir;
@@ -397,7 +481,8 @@ iso_find_entry (struct iso_inode *dir, const char *name, int namelen,
 	block = iso_bmap(dir,f_pos >> bufbits);
 	if (!block) return -1;
 
-	if (iso_breadi(block, 1, data_block) < 0) return -1;
+	if (iso_dev_read(data_block, block * sb.s_blocksize, sb.s_blocksize)
+	    != sb.s_blocksize) return -1;
   
 	while (f_pos < itp->size) {
 		de = (struct iso_directory_record *) (data_block + offset);
@@ -413,7 +498,10 @@ iso_find_entry (struct iso_inode *dir, const char *name, int namelen,
 				 + ISOFS_BLOCK_SIZE);
 			block = iso_bmap(dir,f_pos>>bufbits);
 			if (!block) return -1;
-			if (iso_breadi(block, 1, data_block) < 0) return -1;
+			if (iso_dev_read(data_block,
+					 block * sb.s_blocksize,
+					 sb.s_blocksize)
+				!= sb.s_blocksize) return -1;
   			continue; /* Will kick out if past end of directory */
 		}
 
@@ -433,7 +521,10 @@ iso_find_entry (struct iso_inode *dir, const char *name, int namelen,
 			offset = f_pos & (bufsize - 1);
 			block = iso_bmap(dir,f_pos>>bufbits);
 			if (!block) return -1;
-			if (iso_breadi(block, 1, data_block) < 0) return 0;
+			if (iso_dev_read(data_block,
+					 block * sb.s_blocksize,
+					 sb.s_blocksize)
+				!= sb.s_blocksize) return 0;
 			memcpy((char *)cpnt+frag1, data_block, offset);
 		}
 		
@@ -448,11 +539,6 @@ iso_find_entry (struct iso_inode *dir, const char *name, int namelen,
 
 		if (de->name[0]==1 && de->name_len[0]==1) {
 
-#if 0
-		  printf("Doing .. (%d %d)",
-			 sb.s_firstdatazone,
-			 itp->i_inumber);
-#endif
 		  if((int) sb.s_firstdatazone != itp->inumber)
 		    inode_number = dir->i_backlink;
 		  else
@@ -460,32 +546,30 @@ iso_find_entry (struct iso_inode *dir, const char *name, int namelen,
 		  backlink = 0;
 		}
     
-		dlen = de->name_len[0];
-		dpnt = de->name;
-		/* Now convert the filename in the buffer to lower case */
-#if 0
-		rrflag = get_rock_ridge_filename(de, &dpnt, &dlen, dir);
-#else
-		rrflag = 0;
-#endif
-		if (rrflag) {
-		  if (rrflag == -1) return -1; /* Relocated deep directory */
-		} else {
-		  if(sb.s_mapping == 'n') {
-		    for (i = 0; i < dlen; i++) {
-		      c = dpnt[i];
-		      if (c >= 'A' && c <= 'Z') c |= 0x20;  /* lower case */
-		      if (c == ';' && i == dlen-2 && dpnt[i+1] == '1') {
-			dlen -= 2;
-			break;
+		{
+		  char retname[de->name_len[0] + 1];
+		  dlen = get_rock_ridge_filename(de, retname, dir);
+		  if (dlen) {
+		    strcpy(de->name, retname);
+		  } else {
+		    dlen = de->name_len[0];
+		    if(sb.s_mapping == 'n') {
+		      for (i = 0; i < dlen; i++) {
+			c = de->name[i];
+			if (c >= 'A' && c <= 'Z') c |= 0x20;  /* lower case */
+			if (c == ';' && i == dlen-2
+			    && de->name[i+1] == '1') {
+			  dlen -= 2;
+			  break;
+			}
+			if (c == ';') c = '.';
+			de->name[i] = c;
 		      }
-		      if (c == ';') c = '.';
-		      de->name[i] = c;
+		      /* This allows us to match with and without a trailing
+			 period.  */
+		      if(de->name[dlen-1] == '.' && namelen == dlen-1)
+			dlen--;
 		    }
-		    /* This allows us to match with and without a trailing
-		       period.  */
-		    if(dpnt[dlen-1] == '.' && namelen == dlen-1)
-		      dlen--;
 		  }
 		}
 		/*
@@ -495,14 +579,11 @@ iso_find_entry (struct iso_inode *dir, const char *name, int namelen,
 		if(   !(de->flags[-sb.s_high_sierra] & 5)
 		   || sb.s_unhide == 'y' )
 		{
-		  match = iso_match(namelen,name,dpnt,dlen);
+		  match = iso_match(namelen, name, de->name, dlen);
 		}
 
 		if (cpnt) cpnt = NULL;
 
-#if 0
-		if(rrflag) kfree(dpnt);
-#endif
 		if (match) {
 		  if ((int) inode_number == -1) {
 #if 0
@@ -529,6 +610,10 @@ iso_find_entry (struct iso_inode *dir, const char *name, int namelen,
 		  return 0;
 		}
 	}
+#ifdef DEBUG_ISO
+		  printf("iso_find_entry returning unsuccessfully (ino = %d)\n",
+			 inode_number);
+#endif
 	return -1;
 }
 
@@ -576,9 +661,12 @@ iso_lookup(struct iso_inode *dir, const char *name)
 		} else
 			last++;
 	}
-	if (iso_find_entry(dir, &name[first], last - first, &ino, &ino_back)) {
-		iso_iput(dir);
-		return NULL;
+	{
+		int rv;
+		if ((rv = iso_find_entry(dir, &name[first], last - first, &ino, &ino_back))) {
+			iso_iput(dir);
+			return NULL;
+		}
 	}
 	if (!(result = iso_iget(ino))) {
 		iso_iput(dir);
@@ -615,7 +703,6 @@ int
 iso_read_super (void *data, int silent)
 {
 	static int first_time = 1;
-	unsigned int blocksize_bits;
 	int high_sierra;
 	unsigned int iso_blknum, vol_desc_start;
 	char rock = 'y';
@@ -640,16 +727,8 @@ iso_read_super (void *data, int silent)
 #endif	
 
 	/* set up the block size */
-	blocksize_bits = 0;
-	{
-		int i = 1024;
-		while (i != 1){
-			blocksize_bits++;
-			i >>=1;
-		};
-	}
 	sb.s_blocksize = 1024;
-	sb.s_blocksize_bits = blocksize_bits;
+	sb.s_blocksize_bits = 10;
 
 	sb.s_high_sierra = high_sierra = 0; /* default is iso9660 */
 
@@ -660,8 +739,8 @@ iso_read_super (void *data, int silent)
 #ifdef DEBUG_ISO
 		printf("iso_read_super: iso_blknum=%d\n", iso_blknum);
 #endif
-		if (iso_breadi(iso_blknum << (11 - blocksize_bits), 1,
-			       data_block) < 0)
+		if (iso_dev_read(data_block, iso_blknum * 2048,
+				 sb.s_blocksize) != sb.s_blocksize)
 		{
 			printf("iso_read_super: bread failed, dev "
 			       "iso_blknum %d\n", iso_blknum);
@@ -770,8 +849,6 @@ iso_read_super (void *data, int silent)
 	 * permissions.
 	 */
 	sb.s_mode = S_IRUGO & 0777;
-	sb.s_blocksize = 1024;
-	sb.s_blocksize_bits = blocksize_bits;
 
 	/* return successfully */
 	root_inode = isonum_733 (rootp->extent) << sb.s_log_zone_size;
@@ -783,25 +860,10 @@ int
 iso_bread (int fd, long blkno, long nblks, char * buffer)
 {
 	struct iso_inode *inode;
-	unsigned int block;
 
 	/* find the inode for this file */
 	inode = &inode_table[fd].inode;
-
-	/* do some error checking */
-	if (!inode || !inode->i_first_extent)
-		return 0;
-
-	/* figure out which iso block number(s) we're being asked for */
-	block = iso_bmap(inode, blkno);
-	if (!block)
-		return 0;
-
-	/* now try and read them... */
-	if (iso_breadi(block, nblks, buffer) < 0)
-		return -1;
-
-	return nblks * sb.s_blocksize;
+	return iso_breadi(inode, blkno, nblks, buffer); 
 }
 
 
@@ -843,9 +905,20 @@ iso_map (int fd, long block)
 	return iso_bmap(&inode_table[fd].inode, block) * sb.s_blocksize;
 }
 
-
-size_t
-iso_fsize (int fd)
+#include <linux/stat.h>
+int
+iso_fstat (int fd, struct stat * buf)
 {
-	return inode_table[fd].size;
+	struct inode_table_entry * ip = inode_table + fd;
+
+	if (fd >= MAX_OPEN_FILES)
+		return -1;
+
+	memset(buf, 0, sizeof(struct stat));
+	/* fill in relevant fields */
+	buf->st_ino   = ip->inumber;
+	buf->st_mode  = ip->mode;
+	buf->st_nlink = ip->nlink;
+	buf->st_size  = ip->size;
+	return 0;
 }
